import logging
import os
import subprocess

from contextlib import contextmanager

import psycopg2
from psycopg2 import sql

try:
    import metabase
    from const import PG_ODOO_USER, PG_PASS, PG_USER
except ImportError:
    from . import metabase
    from .const import PG_ODOO_USER, PG_PASS, PG_USER

try:
    from odoo import tools
    config = tools.config
except ImportError:
    # Odoo-less usage
    config = {
        "db_host": "localhost",
        "db_template": "template0",
    }


_logger = logging.getLogger(__name__)

db_host = config["db_host"] or "localhost"
pg_env = dict(os.environ, PGPASSWORD=PG_PASS) if PG_PASS else dict(os.environ)
pg_cnx = ["-h", db_host, "-U", PG_USER] if PG_PASS else []


@contextmanager
def cursor(dbname):
    dsn = {
        "host": os.environ.get("PGHOST", db_host),
        "dbname": dbname,
    }
    if PG_PASS:
        dsn["password"] = PG_PASS
        dsn["user"] = PG_USER

    conn = psycopg2.connect(**dsn)
    conn.autocommit = True

    cr = conn.cursor()

    try:
        yield cr
    except Exception:
        raise
    finally:
        cr.close()
        conn.close()


def _do_db_mgmt(db_name, query, action):
    """Copied from odoo.service.db because it does not work
    for databases that are not visible in db_list anymore,
    but we may still need to drop them"""
    with cursor("postgres") as cr:
        # Try to terminate all other connections that might prevent
        # dropping the database
        try:
            cr.execute("""
                SELECT pg_terminate_backend(pid)
                  FROM pg_stat_activity
                 WHERE datname = %s
                   AND pid != pg_backend_pid()
            """, (db_name,))
        except Exception:
            _logger.exception("Failed to drop connections: db=%s", db_name)

        try:
            cr.execute(query)
            _logger.info("%s: %s", action, db_name)
        except Exception:
            _logger.error("%s: %s failed", action, db_name)
            raise Exception("Couldn't %s %s" % (action, db_name))


def createdb(dbname, template=None):
    """
    Creates a new PostgreSQL database. Optionally, a template database can be
    provided.
    """
    dbname = dbname.lower()
    chosen_template = template or config["db_template"]
    cmd = ["createdb"] + pg_cnx + ["--encoding", "unicode", "--template", chosen_template, dbname]
    with open(os.devnull, 'ab') as dn:
        res = subprocess.call(cmd, env=pg_env, stdout=dn, stderr=subprocess.STDOUT)
        if res:
            _logger.error("Could not create database %s" % dbname)
            raise Exception("Could not create database %s" % dbname)


def create_empty_database(dbname):
    """
    Uses a spare if available. Creates a new database otherwise.
    """
    dbname = dbname.lower()
    try:
        spare = metabase.spare_use()
        rename(spare, dbname)
    except Exception:
        createdb(dbname)
    set_public_schema_privileges(dbname)


def create_extension(dbname, extension):
    """
    Connects as database owner and installs extension into dbname.
    """
    with cursor(dbname) as cr:
        query = "CREATE EXTENSION IF NOT EXISTS {}".format(extension)
        cr.execute(query)


def drop(dbname):
    query = 'DROP DATABASE IF EXISTS "%s"' % dbname
    _do_db_mgmt(dbname, query, "drop")


def dump_restore(db_from, db_to):
    """
    Dump and restore the database db_from to db_to.
    Do not use an intermediary file but pipe the result of pg_dump directly to pg_restore. No
    compression to avoid any CPU overhead.

    :param db_from: source database name
    :param db_to: destination database name
    """

    db_from = db_from.lower()
    db_to = db_to.lower()

    create_empty_database(db_to)

    cmd_dump = ["pg_dump"] + pg_cnx + ["--format=c", "--compress=0", db_from]
    cmd_restore = ["pg_restore"] + pg_cnx + ["--dbname=" + db_to]

    _logger.info("Dump and restore database %s to %s start", db_from, db_to)
    # Pipe the dump to the restore command
    with subprocess.Popen(cmd_dump, stdout=subprocess.PIPE, env=pg_env) as dump:
        with subprocess.Popen(cmd_restore, stdin=dump.stdout, env=pg_env) as restore:
            dump.stdout.close()
            rest_out, rest_err = restore.communicate()

    if dump.returncode != 0 or restore.returncode != 0:
        err_msg = "Dump and/or restore failed! "
        if dump.returncode:
            err_msg += "Dump code: {}".format(dump.returncode)
        if restore.returncode:
            err_msg += "Restore code: {}".format(restore.returncode)
            err_msg += "Restore stdout: {}".format(rest_out)
            err_msg += "Restore stderr: {}".format(rest_err)
        _logger.error(err_msg)
    else:
        _logger.info("Dump and restore database %s to %s success", db_from, db_to)

def dump(dbname, filename, format="c", no_owner=False):
    dbname = dbname.lower()
    assert format in ["c", "d", "t", "p"]
    cmd = ["pg_dump"] + pg_cnx + ["--format=" + format, "--file=" + filename]
    if no_owner:
        cmd.append("--no-owner")
    cmd.append(dbname)

    res = subprocess.call(cmd, env=pg_env)
    if res:
        _logger.exception("Could not dump database %s" % dbname)
        raise Exception("Could not dump database %s" % dbname)


def rename(old, new):
    query = 'ALTER DATABASE "%s" RENAME TO "%s"' % (old, new)
    _do_db_mgmt(old, query, "rename")


def restore(dbname, filename, format="custom"):
    """
    Restores dbname from filename where filename is an archive created by
    pg_dump in one of the non-plain-text formats.

    Use format="plain" to restore from plain-text SQL script file.
    """
    dbname = dbname.lower()
    create_empty_database(dbname)

    if format == "custom":
        cmd = ["pg_restore"] + pg_cnx + ["--dbname=" + dbname]
        cmd.append(filename)
    elif format == "plain":
        cmd = ["psql"] + pg_cnx + ["-q", "-f", filename, dbname]
    else:
        raise ValueError("Invalid format: %s" % format)

    with open(os.devnull, 'ab') as dn:
        try:
            subprocess.check_call(cmd, env=pg_env, stdout=dn, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError:
            _logger.exception("Could not restore database %s" % dbname)
            raise Exception("Could not restore database %s" % dbname)


def set_public_schema_privileges(dbname):
    """
    Grant the CREATE privilege on the public schema to the odoo role.
    See https://www.postgresql.org/docs/15/ddl-schemas.html#DDL-SCHEMAS-PATTERNS
    """
    pg_version_query = "SHOW server_version_num"
    create_priv_query = sql.SQL("GRANT CREATE ON SCHEMA public TO {}").format(
        sql.Identifier(PG_ODOO_USER)
    )
    with cursor(dbname) as cr:
        cr.execute(pg_version_query)
        pg_version = cr.fetchone()[0]
        if int(pg_version) >= 150000:
            cr.execute(create_priv_query)


def update_template_lru(template_name):
    """
    Update the (trial database) template LRU. The given template_name last used
    timestamp is set to now.
    The last used timestamp is stored as the database template "stat_reset"
    timestamp. Its update is done with pg_stat_reset().
    """
    try:
        with cursor(template_name) as cr:
            try:
                timeout_ms = 1000
                timeout_query = "SET statement_timeout = %s"
                cr.execute(timeout_query, (timeout_ms,))
                cr.execute("SELECT pg_stat_reset()")
                _logger.info("Template LRU: updated stats_reset. db=%s", template_name)
            except Exception:
                _logger.exception(
                    "Template LRU: failed to update stats_reset, transaction failed. db=%s",
                    template_name,
                )
    except Exception:
        _logger.exception(
            "Template LRU: failed to update stats_reset, "
            "connection closed or couldn't be established. db=%s",
            template_name,
        )
